#ifndef __ALIGN_H__
#define __ALIGN_H__

#include "image_types.h"
#include "kdtree.h"
#include <vector>

namespace s3d_merge
{

enum AlignmentParameters
{
    ALIGN_ANGLE=0,
    ALIGN_TRANSLATE_X,
    ALIGN_TRANSLATE_Y,
    ALIGN_NUM_PARAMETERS
};

typedef spectral::Point<double,  2> Point2D;
typedef spectral::KDTree<double, 2> KDTree2D;

typedef spectral::Point<double,  3> Point3D;
typedef spectral::KDTree<double, 3> KDTree3D;

typedef spectral::Point<double,  6> PatternNode;
typedef spectral::KDTree<double, 6> PatternTree;


class StarRef
{
public:
    StarRef()
        : m_value(0)
        , m_pixels(0)
    {
        m_point[0] = m_point[1] = 0.0;
    }

    StarRef(double x, double y, double value, size_t pixels)
        : m_value(value)
        , m_pixels(pixels)
    {
        m_point[0] = x;
        m_point[1] = y;
    }

    StarRef(const StarRef &other)
        : m_point(other.GetPoint())
        , m_value(other.GetBrightness())
        , m_pixels(other.GetPixelCount())
    {
    }

    const StarRef &operator=(const StarRef &other)
    {
        m_point = other.GetPoint();
        m_value = other.GetBrightness();
        m_pixels = other.GetPixelCount();
        return *this;
    }

    const Point2D &GetPoint(void) const { return m_point; }

    double GetBrightness(void) const { return m_value; }

    size_t GetPixelCount(void) const { return m_pixels; }
private:
    Point2D m_point;
    double m_value;
    size_t m_pixels;
};

bool operator<(const s3d_merge::StarRef &a, const s3d_merge::StarRef &b);

// Locate candidate stars in an image
void FindStars(const Image16 *img, size_t max_stars, std::vector<StarRef> &stars);


// 2D pattern generation
PatternNode 
GeneratePatternNode2D(const std::vector<Point2D> &points,
                      const KDTree2D *tree,
                      size_t index);

void
GeneratePatternNodes2D(const std::vector<Point2D> &points,
                       const KDTree2D *tree,
                       std::vector<PatternNode> &out);

// 3D pattern generation
PatternNode
GeneratePatternNode3D(const std::vector<Point3D> &points,
                      const KDTree3D *tree,
                      size_t index);

void
GeneratePatternNodes3D(const std::vector<Point3D> &points,
                       const KDTree3D *tree,
                       std::vector<PatternNode> &out);

// 2D to 2D star fit.
bool
FitStars(const PatternTree *ref_pattern_tree,
         const std::vector<PatternNode> &ref_patterns,
         const KDTree2D *ref_tree,
         const std::vector<Point2D> &ref_points,
         const std::vector<Point2D> &cur_points,
         double *alignment);

}

// Sort in descending order of magnitude

#endif
