#include <math.h>
#include <stdlib.h>
#include <stdint.h>
//#include "utils.h"
#include "sampler.h"
#include "image.h"


#ifndef PI
#define PI 3.14159265
#endif

// un normalised sinc
double sincu(double x)
{
    if(fabs(x) < 1.0e-10)
    {
        return 1.0;
    }
    else
    {
        return sin(x) / x;
    }
}

// normalised sinc
double sincn(double x)
{
    return sincu(x * PI);
}

// Lancszos function.
double lanczos(double x, double a)
{
    if(fabs(x) < a)
    {
        return sincn(x)*sincn(x/a);
    }
    else
    {
        return 0;
    }
}


////////////////////////////////////////////////////////////////////////////////
const float *
s3d_merge::ImageSampler::GetPixel(const s3d_merge::FloatImage *image, 
                                 int32_t x, int32_t y) const
{
    const float *result(NULL);

    if(image)
    {
        uint32_t width, height;
        bool x_valid(true), y_valid(true);
        width = image->get_width();
        height = image->get_height();

        if((x < 0) || (x >= (int32_t)width))
        {
            switch(m_horizontalWrap)
            {
            case BACKGROUND:
                x_valid = false;
                break;
            case CLAMP:
                if(x < 0)
                {
                    x = 0;
                }
                else
                {
                    x = (int)width - 1;
                }
                break;
            case REFLECT:
                x = FloatImage::constrain_reflect(x, width);
                break;
            case WRAP:
                x = FloatImage::constrain_wrap(x, width);
                break;
            }
        }
        if((y < 0) || (y >= (int32_t)height))
        {
            switch(m_verticalWrap)
            {
            case BACKGROUND:
                y_valid = false;
                break;
            case CLAMP:
                if(y < 0)
                {
                    y = 0;
                }
                else
                {
                    y = (int)height - 1;
                }
                break;
            case REFLECT:
                y = FloatImage::constrain_reflect(y, height);
                break;
            case WRAP:
                y = FloatImage::constrain_wrap(y, height);
                break;
            }
        }
        if(x_valid && y_valid)
        {
            result = image->get_pixel(x, y);
        }
    }
    return result;
}

////////////////////////////////////////////////////////////////////////////////
const float *
s3d_merge::ImageSampler::GetPixel(const s3d_merge::FloatImage *image, 
                                 float x, float y) const
{
    int32_t ix, iy;

    ix = round(x);
    iy = round(y);
	
    return GetPixel(image, ix, iy);
}

////////////////////////////////////////////////////////////////////////////////
void 
s3d_merge::NearestNeighbourSampler::SampleImage(const FloatImage *image, 
                                               float x, float y,
                                               const float *bg,
                                               float *result) const
{
    if(image && result)
    {
        uint32_t channels = image->get_channels();
        const float *p = GetPixel(image, x, y);

        if(p)
        {
            for(uint32_t i=0;i<channels;i++)
            {
                result[i] = p[i];
            }
        }
        else
        {
            if(bg)
            {
                for(uint32_t i=0;i<channels;i++)
                {
                    result[i] = bg[i];
                }
            }
            else
            {
                for(uint32_t i=0;i<channels;i++)
                {
                    result[i] = 0.0;
                }
            }
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
void
s3d_merge::LanczosSampler::SampleImage(const FloatImage *image, 
                                      float x, float y,
                                      const float *bg,
                                      float *result) const
{
    if(image && result)
    {
        int32_t min_x, max_x, min_y, max_y, ix, iy;
        uint32_t channels;
        double total_weight(0);

        min_x = floor(x - m_a);
        min_y = floor(y - m_a);
        max_x = ceil(x + m_a);
        max_y = ceil(y + m_a);

/*
        width = image->get_width();
        height = image->get_height();
*/
        channels = image->get_channels();

        for(uint32_t i=0;i<channels;i++)
        {
            result[i] = 0;
        }

        for(iy = min_y; iy <= max_y; iy++)
        {
            double ly;
            ly = lanczos(y - float(iy), m_a);

            for(ix = min_x; ix <= max_x; ix++)
            {
                double weight;
                weight = lanczos(x - float(ix), m_a) * ly;

                if(fabs(weight) > 1.0e-10)
                {
                    total_weight+= weight;
                    if(const float *pixel = GetPixel(image, ix, iy))
                    {
                        for(uint32_t i=0;i<channels;i++)
                        {
                            result[i]+= pixel[i] * weight;
                        }
                    }
                    else
                    {
                        if(bg)
                        {
                            for(uint32_t i=0;i<channels;i++)
                            {
                                result[i]+= bg[i] * weight;
                            }
                        }
                    }
                }
            }
        }

        if(fabs(total_weight) > 1.0e-10)
        {
            for(uint32_t i=0;i<channels;i++)
            {
                result[i]/= total_weight;
                if(result[i] < 0.0) result[i] = 0.0;
            }
        }
        else if(bg)
        {
            for(uint32_t i=0;i<channels;i++)
            {
                result[i] = bg[i];
            }
        }
        else
        {
            for(uint32_t i=0;i<channels;i++)
            {
                result[i] = 0.0;
            }
        }
    }
}


